# Copyright © 2022 Arrikto Inc.  All Rights Reserved.

"""Kale SDK.

This script uses an ML pipeline to train and serve an SKLearn Model.
"""

import os
import joblib

from typing import Tuple, NamedTuple

from sklearn.feature_extraction import text
from sklearn.naive_bayes import MultinomialNB
from sklearn.datasets import fetch_20newsgroups
from sklearn.model_selection import train_test_split
from sklearn.feature_extraction.text import TfidfVectorizer

from kale.ml import Signature
from kale.types import MarshalData
from kale.sdk import pipeline, step
from kale.common import mlmdutils, artifacts


ASSETS_PATH = "/home/jovyan/transformer_package/"


@step(name="data_loading")
def load_split_dataset() -> Tuple[MarshalData, MarshalData]:
    """Fetch 20newgroup dataset."""
    # load the data
    newsgroups_dataset = fetch_20newsgroups(random_state=42)
    x = newsgroups_dataset.data
    y = newsgroups_dataset.target

    x, _, y, _ = train_test_split(x, y, test_size=.2, random_state=42)

    return x, y


@step(name="data_preprocess")
def preprocess(x: MarshalData) -> Tuple[MarshalData, int]:
    """Preprocess the input data."""
    # get stopwords
    stop_words = text.ENGLISH_STOP_WORDS
    # TF-IDF vectors
    vectorizer = TfidfVectorizer(stop_words=stop_words)
    x_processed = vectorizer.fit_transform(x)

    with open(os.path.join(ASSETS_PATH, "vectorizer.joblib"), "wb") as f:
        joblib.dump(vectorizer, f)

    # create and submit a Transformer artifact
    mlmd = mlmdutils.get_mlmd_instance()

    transformer_artifact = artifacts.Transformer(
        name="Vectorizer",
        transformer_dir=ASSETS_PATH,
        module_name="transformer",
        class_name="Transformer",
        is_stateful=True
    ).submit_artifact()

    mlmd.link_artifact_as_output(transformer_artifact.id)

    return x_processed, transformer_artifact.id


@step(name="model_training")
def train(x: MarshalData,
          y: MarshalData) -> NamedTuple("outs", [("model", MarshalData)]):
    """Train a MultinomialNB model."""
    classifier = MultinomialNB(alpha=.01)
    model = classifier.fit(x, y)
    return model


@step(name="register_model")
def register_model(model: MarshalData, x: MarshalData, y: MarshalData) -> int:
    mlmd = mlmdutils.get_mlmd_instance()

    signature = Signature(
        input_size=[1] + list(x[0].shape),
        output_size=[1] + list(y[0].shape),
        input_dtype=x.dtype,
        output_dtype=y.dtype)

    model_artifact = artifacts.SklearnModel(
        model=model,
        description="A simple MultinomialNB model",
        version="1.0.0",
        author="Kale",
        signature=signature,
        tags={"app": "sklearn-tutorial"}).submit_artifact()

    mlmd.link_artifact_as_output(model_artifact.id)
    return model_artifact.id


@pipeline(name="classification", experiment="sklearn-tutorial")
def ml_pipeline():
    """Run the ML pipeline."""
    x, y = load_split_dataset()
    x_processed, _ = preprocess(x)
    model = train(x_processed, y)
    register_model(model, x_processed, y)


if __name__ == "__main__":
    ml_pipeline()
